
from basics import *
from numpy.core.numerictypes import *
import numpy
from FITS import *
from VTKHelperFunctions import *
from pgplot import *

import ZODB, ZODB.FileStorage, transaction
import persistent
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree

from GalPhot import *

def CleanContours(contours,area=4):
    cleaned = []
    for contour in contours:
        x1,y1 = transpose(asarray(contour)[::,::,0])
        x2,y2 = transpose(asarray(contour)[::,::,1])
        a = 0.5*sum(x1*y2-y1*x2)
        if a > area:
           cleaned.append(contour)
    return cleaned

def DoContour( image,n1,level,th,area,smo):
    F = FITS(image,readonly=0)
    frame = F.data()
    naxis1 = F["naxis1"]
    naxis2 = F["naxis2"]
    sigma = F["bnoise"]
    du = compress(frame.ravel(),frame.ravel())
    sigma = 1.49*median(abs(du))
    F["bnoise0"] = sigma
    sframe = VTKGauss(frame,smo,numret=1)
    du = compress(frame.ravel(),sframe.ravel())
    sigma = 1.49*median(abs(du))
    F.modkey("bnoise",sigma)
    F.close()

    print "Smooth=",smo
    print "Noise=",sigma

    contours = {}
    masks = {}

    print "Contouring %.1f" % (level),
    area1=max([area*(th/level)**2,4])
    print image,level,level*sigma,area1
    t0 = time.time()
    contours[level] = VTKGetContours(sframe,level*sigma,minsides=2,area=area1,verb=1,pad=0)
    t1 = time.time()
    print level,len(contours[level])
#    masks[level] = zeros(frame.shape,float64)
#    masks[level] = VTKDrawContour(masks[level],contours[level],color=level*sigma,fill=level*sigma,verb=0).astype(float32)
#    t2 = time.time()
#    print
#    qdump("mask%04d_%04d.fits" % (n1,level),masks[level])
    print "Timing = %.1f" % (t1-t0)
    sys.stdout.flush()
    zdb = "contours%04d_%04d.zodb" % (n1,level)
    WriteContours(zdb,contours[level])

def WriteContours(zdb,contours):
    try:
       if os.path.exists(zdb): os.remove(zdb)
       if os.path.exists(zdb+".tmp"): os.remove(zdb+".tmp")
       if os.path.exists(zdb+".index"): os.remove(zdb+".index")
       if os.path.exists(zdb+".lock"): os.remove(zdb+".lock")
    except:
       print "Odd, we couldn't remove old zdb stuff?"

    storage = ZODB.FileStorage.FileStorage(zdb)
    print "Writing to",zdb
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    dbroot["Contours"] = OOBTree()
    if isinstance(contours,ListType):
       items = range(len(contours))
    else:
       items = contours.keys()
       items.sort()
    for j in items:
        dbroot["Contours"][j] = contours[j]
    transaction.commit()
    conn.cacheMinimize()

    conn.close()
    db.close()
    storage.close()

def ReadFamilies(zdb,verb=0):
    storage = ZODB.FileStorage.FileStorage(zdb,read_only=True)
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    contours = {}
    ck = list(dbroot["Contours"].keys())
    print "Reading %s(%d)..." % (zdb,len(ck)),
    sys.stdout.flush()
    for j in ck:
        contours[j] = copy.copy(dbroot["Contours"][j])
        if verb:
           print j,
           sys.stdout.flush()
    conn.close()
    db.close()
    sys.stdout.flush()
    storage.close()
    print "Closed!"
    sys.stdout.flush()
    return contours


def ReadContours(zdb,verb=0,noclose=False):
    storage = ZODB.FileStorage.FileStorage(zdb,read_only=True)
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    contours = {}
    ck = list(dbroot["Contours"].keys())
    print "Reading %s(%d)..." % (zdb,len(ck)),
    sys.stdout.flush()
    for j in ck:
        contours[j] = copy.copy(dbroot["Contours"][j])
        if verb:
           print j,
           sys.stdout.flush()
    if noclose:
       return contours,conn,db,storage
    else:
       conn.close()
       db.close()
       storage.close()
       print "Closed!"
       sys.stdout.flush()
       return contours

def AddNoise(image):
    F = FITS(image,readonly=0)
    frame = F.data()
    du = compress(frame.ravel(),frame.ravel())
    sigma = 1.49*median(abs(du))
    F.modkey("bnoise",sigma)
    F.close()
    print "Noise in %s is %f" % (image,sigma)

def GetContours( image,n1,n0,levels):
    F = FITS(image)
    naxis1 = F["naxis1"]
    naxis2 = F["naxis2"]
    sigma = F["bnoise"]
    F.close()

    oframe = zeros((naxis2,naxis1),float64)
    print "Noise=",sigma

    contours = {}
    ellipses = {}
    for level in levels:
        zdb = "contours%04d_%04d.zodb" % (n1,level)
        contours[level] = ReadContours(zdb,verb=0)
        ellipses[level] = {}
        for j in contours[level]:
           contours[level][j] = Contour(contours[level][j],level*sigma)
           (xb,yb),(xe,ye) = transpose(contours[level][j].contour)
           wb = level*sigma*ones(xb.shape)
           ellipses[level][j] = Ellipse(xb,yb,wb,ns=0)

    if 0:
      for l,level in enumerate(levels):
          print "Fit ellipses %d. Try to zero out the overlaps." % (l)
          for j in contours[level]:
              contour = contours[level][j]
              dmin = 0.25*sqrt(contour.hei**2+contour.wid**2)
              dmax = 4*sqrt(contour.hei**2+contour.wid**2)
              (xb,yb),(xe,ye) = transpose(contour.contour)
              wb = level*sigma*ones(xb.shape)
              if level != levels[-1]:
#                 ellipses[level][j] = Ellipse(xb,yb,wb)
                 for l2 in levels[l+1:]:
                    for j2 in contours[l2]:
                        c1 = contours[l2][j2]
                        d = sqrt((c1.mx-contour.mx)**2 + (c1.my-contour.my)**2)
                        if dmin <= d <= dmax:
                           c2 = c1.multiply(2)
                           inx = greater_equal(c2.mx,contour.ex[0])*less_equal(c2.mx,contour.ex[1])
                           iny = greater_equal(c2.my,contour.ey[0])*less_equal(c2.my,contour.ey[1])
                           if (inx*iny).any():
                              t = Interiors(xb,yb,c2.contour,verb=0)
                              interior = less(abs(t-2*pi),1e-4)
                              wb = wb*logical_not(interior)
#                              if interior.any():
#                                 print level,j,sum(inx*iny),len(wb),sum(greater(wb,0))
              ellipses[level][j] = Ellipse(xb,yb,wb)
#    exit()

    return contours, ellipses

def CatalogFamilies(image,suff,noise,zdb,ztable,ext=1.2,next=4):
    print "Reading",image
    sigma = image.replace(suff,noise)
    print "And reading",sigma
    catalog = image.replace(".fits",".cat")
    print "Generating",catalog
    print catalog
    F = FITS(image)
    naxis1 = F["naxis1"]
    naxis2 = F["naxis2"]
    data = F.data()
    filt = F["filter"]
    if filt == "N/A" or "CLEAR" in filt:
       filt = F["filter1"]
    if filt == "N/A" or "CLEAR" in filt:
       filt = F["filter2"]
    filt = filt.strip()
    F.close()
    F = FITS(sigma)
    sigma = F.data()
    F.close()
    vari = power(sigma,2)
    ztable = readtable(ztable,header=0,tcols=["name","filter"])
    zpt = [zpt for f,zpt in zip(ztable["filter"],ztable["zpt"]) if f == filt][0]
    print "Got zeropoint of",zpt

    parents,conn,db,storage = ReadContours(zdb,verb=0,noclose=True)
    print

    print "Add up fluxes within ellipses..."
    ws = parents.keys()
    myo = []
    for w in ws:
        uw = parents[w].keys()
        mys = [parents[w][v][1].my for v in uw]
        ars = [parents[w][v][1].area for v in uw]
        myo.append( mys[argmin(ars)])
    order = argsort(myo)
    ws = [ws[j] for j in order]

    cout = open(catalog,"w")
    for j,w in enumerate(ws):
        if j and not (j%(len(ws)/80+1)):
           sys.stdout.write(".")
           sys.stdout.flush()
           conn.cacheMinimize()
        uw = parents[w].keys()
        ars = [parents[w][v][1].area for v in uw]
        uw = [uw[j] for j in argsort(ars)]
        parents[w][uw[0]][1].children = uw
        counts = {}
        errors = {}
        mags = {}
        errs = {}
        ab = [parents[w][v][1].ab for v in uw]
        ba = [(parents[w][v][1].b/parents[w][v][1].a) for v in uw]
        if len(ab)>1: arat = median(divz(ab[1:],ab[:-1]))
        else: arat = 1.0
        abwt1 = []
        abwt2 = []
        arwt1 = []
        arwt2 = []
        bawt1 = []
        bawt2 = []
        for j,v in enumerate(uw):
            xp,yp,zp = parents[w][v][1].contour.pixels(data)
            zp2 = parents[w][v][1].contour.pixels(vari)[2]
            ab1 = parents[w][v][1].ab
            counts[ab1] = sum(zp)
            errors[ab1] = sqrt(sum(zp2))
            if j:
               dcounts = max([(counts[ab1]-counts[ab[j-1]]),0.0])
               abwt1.append(dcounts* parents[w][v][1].ab)
               abwt2.append(dcounts)
               arwt1.append(dcounts* parents[w][v][1].ab/ab[j-1])
               arwt2.append(dcounts)
               bawt1.append(dcounts* parents[w][v][1].b/ parents[w][v][1].a)
               bawt2.append(dcounts)
            if counts[ab1] > 0:
               mags[ab1] = -2.5*log10(counts[ab1]) + zpt
               errs[ab1] = (2.5/log(10))*divz(errors[ab1],counts[ab1])
            else:
               mags[ab1] = -99.999
               errs[ab1] = -9.999
            transaction.abort()
        if len(uw) > 1:
           last = parents[w][uw[-1]][1]
           for n in range(next):
               xp,yp,zp = last.contour.multiply(pow(ext,1+n)).pixels(data)
               zp2 = last.contour.multiply(pow(ext,1+n)).pixels(vari)[2]
               ab1 = last.ab*pow(ext,1+n)
               ab.append(ab1)
               counts[ab1] = sum(zp)
               errors[ab1] = sqrt(sum(zp2))
               if counts[ab1] > 0:
                  mags[ab1] = -2.5*log10(counts[ab1]) + zpt
                  errs[ab1] = (2.5/log(10))*divz(errors[ab1],counts[ab1])
               else:
                  mags[ab1] = -99.999
                  errs[ab1] = -9.999
           
        v = uw[0]
        mab = divz(sum(abwt1), sum(abwt2))
        mar = divz(sum(arwt1), sum(arwt2))
        mba = divz(sum(bawt1), sum(bawt2))
        record = ["%20s"%(w)]
        record += ["%10.4f %10.4f" % (parents[w][v][1].mx,parents[w][v][1].my)]
        record += ["%10.4f %10.4f %10.4f" % (mar,mab,mba)]
        record += ["%10.4f %10.4f %10.4f %10.4f" % (ab1,ba1,mags[ab1],errs[ab1]) for ab1,ba1 in zip(ab,ba)]
        line = " ".join(record)+"\n"
        cout.write(line)
        cout.flush()
    cout.close()

    print
    conn.close()
    db.close()
    storage.close()
    print "ZODB Closed!"
    sys.stdout.flush()

    return parents

def ContourFamily(image,image2,n1,n0,levels,efit,all):

    o2 = lambda x,n: x.replace("o%04d.fits"%(n1),"o%04d.fits"%(n))
    o3 = lambda x,n: x.replace("o%04d.fits"%(n1),"m%04drms.fits"%(n))
    d2 = lambda n: "d%04d.fits" % (n)

    ne = int(log(1.*n1/n0)/log(2))
    ns = n0*power(2,arange(ne+1))

    print "Image=",image
#    uimage = o2(image,2*n1)
    uimage = o2(image,ns[-1])
    uimage2 = o3(image,ns[-1])
    print "Reading",uimage
    F = FITS(uimage)
    naxis1 = F["naxis1"]
    naxis2 = F["naxis2"]
    sigma = F["bnoise0"]
    data = F.data()
    F.close()
    F = FITS(uimage2)
    data2 = F.data()
    sigma = median(compress(data.ravel(),data2.ravel()))
    F.close()
    print "Got sigma=",sigma

    contours = {}
    ellipses = {}

    if all:
       for j2,n2 in enumerate(ns):
           contours[n2], ellipses[n2] = GetContours( d2(n2),n2,n0,levels[j2:])
    else:
       contours[n1], ellipses[n1] = GetContours( image2,n1,n0,levels)


    conall = {}
    ellall = {}
    for n2 in ns:
        if n2 in contours:
           ne = int(log(1.*n2/n0)/log(2))
           for level in levels:
               if level in contours[n2]:
                  for j in contours[n2][level]:
                      if not ellipses[n2][level][j].fail:
                         w = "%04d%05d%05d" % (n2,level,j)
                         conall[w] = contours[n2][level][j]
                         ellall[w] = ellipses[n2][level][j]

    ws = ellall.keys()
    print len(ws)
    ws = [w for w in ws if not ellall[w].fail]
    print len(ws)

    ab = array([ellall[w].ab for w in ws])
    area = array([conall[w].area for w in ws])
    print "Quartiles of sqrt(ab)=",quartiles(ab)
    print "Extrema of sqrt(ab)=",extrema(ab)
    print "Quartiles of area=",quartiles(area)
    print "Extrema of area=",extrema(area)
    order = argsort(area)[::-1]
    ws = [ws[j] for j in order]
#    ws = [w for w in ws if 2400<conall[w].mx<2600 and 2400<conall[w].my<2600]
#    ws = [w for w in ws if 2200<conall[w].mx<2800 and 2200<conall[w].my<2800]
#    ws = [w for w in ws if 2000<conall[w].mx<3000 and 2000<conall[w].my<3000]
#    ws = [w for w in ws if 1000<conall[w].mx<2000 and 2000<conall[w].my<3000]

    if 1:
       print "Get enclosed counts for",len(ws)
       x0s,x1s,y0s,y1s,boxes,areas = [],[],[],[],[],[]
       for j,w in enumerate(ws):
           x0,x1 = int(conall[w].ex[0]), int(conall[w].ex[1]+1)
           y0,y1 = int(conall[w].ey[0]), int(conall[w].ey[1]+1)
           box = (x1-x0)*(y1-y0)
           areas.append(conall[w].area)
           x0s.append(x0)
           x1s.append(x1)
           y0s.append(y0)
           y1s.append(y1)
           boxes.append(box)
       x0s,x1s,y0s,y1s,boxes,areas = asarray( [x0s,x1s,y0s,y1s,boxes,areas])
       print extrema(x0s), extrema(x1s), extrema(x1s-x0s)
       print extrema(y0s), extrema(y1s), extrema(y1s-y0s)
       print extrema(boxes),quartiles(boxes)
       print extrema(areas),quartiles(areas)

       t0 = time.time()
       for j,w in enumerate(ws):
           if j and not (j%(len(ws)/160+1)):
              sys.stdout.write(".")
              sys.stdout.flush()
           x0,x1 = int(conall[w].ex[0]), int(conall[w].ex[1]+1)
           y0,y1 = int(conall[w].ey[0]), int(conall[w].ey[1]+1)
           box = (y1-y0)*(x1-x0)
#           print "%s %9.1f %4d %4d %4d %4d %6d %7.4f" % (w,conall[w].area,x0,x1,y0,y1,box,1.0*conall[w].area/box,)
#           sys.stdout.flush()
           xpixels, ypixels, zpixels = conall[w].pixels(data)
           wpixels = divz(zpixels*greater(zpixels,0),sqrt((getsqrt(zpixels)**2+sigma**2)))
           upixels = greater(wpixels,0)
           xpixels,ypixels,wpixels,zpixels = compress(upixels, \
                   [xpixels,ypixels,wpixels,zpixels], 1)
           conall[w].counts = sum(zpixels)
           conall[w].ix = xpixels
           conall[w].iy = ypixels
           conall[w].iz = zpixels
           conall[w].flux = divz(conall[w].counts, conall[w].area)

#           if len(wpixels) > max([0.25*conall[w].area,pi]):
#             e2 = Ellipse(xpixels,ypixels,wpixels,ns=0,verb=0)
##             if diff(extrema(e2.ellx))[0]>0.75*(naxis1-2) or \
##                      diff(extrema(e2.elly))[0]>0.75*(naxis2-2):
##                e2.fail = True
#             if e2.fail:
#                conall[w].counts = 0.0
##             else:
##                print w, len(xpixels), conall[w].area, conall[w].ex, extrema(e2.ellx), conall[w].ey, extrema(e2.elly)
##                conall[w].counts = sum(zpixels)
##                val = mean(zpixels)-3*sigma
##                mag = divz(val*conall[w].area,conall[w].counts)
##           print conall[w].counts
##           sys.stdout.flush()
##           raw_input()

       t1 = time.time()
       print
       print "Time to get enclosed counts =",t1-t0

    print len(ws)
    ws = [w for w in ws if not ellall[w].fail and conall[w].counts>0]
    print len(ws)
    
    laps = {}
    print "Make families!",len(ws)
    used = {}
    for j,w in enumerate(ws[:-1]):
        if j and not (j%(len(ws)/80+1)): sys.stdout.write("."); sys.stdout.flush()
        lap1 = GetOverlaps(w,ws[j+1:],conall,ellall)
        used.setdefault(w,w)
        if lap1:
           u = [w2 for w2 in lap1 if w2 in used]
           if 0 and len(u):
              wa = list(Set([used[w2] for w2 in u]))
              w1 = wa[argsort([conall[w2].area for w2 in wa])[-1]]
              [laps[w1].setdefault(w2) for w2 in lap1]
              [used.__setitem__(w2,w1) for w2 in lap1]
           else:
              laps[w] = lap1
              [used.__setitem__(w2,w) for w2 in lap1]
    print
    print "Number of families",len(laps)
    
    print "Prune families from",len(laps)
    popping = True
    pn = 0
    pruned = {}
    while popping:
       popping = False
       pn += 1
       if pn and not (pn%100):
          sys.stdout.write(".")
          sys.stdout.flush()
       for w in ws:
           if w in laps:
              for w2 in laps[w].keys():
                  if w!=w2 and w2 not in pruned and w2 in laps:
                     [laps[w].setdefault(w3) for w3 in laps[w2]]
                     [used.__setitem__(w3,w) for w3 in laps[w2]]
                     pruned.setdefault(w2)
                     laps.pop(w2)
                     popping = True
                     break
           if popping: break
    
    print "Number of families left",len(laps)
    print Set([len(laps[w]) for w in ws if w in laps])

    sp = {}
    [sp.__setitem__(l,sp.get(l,0)+1) for k in laps for l in laps[k]]
    print Set(sp.values())

    rats = []

    for w in laps:
        if len(laps[w])>2:
           vs = laps[w].keys()
           vs = [vs[j] for j in argsort([ellall[v].ab for v in vs])]
           evs = [ellall[v].v for v in vs]
           eabs = [ellall[v].ab for v in vs]
           rats1 = divz(eabs[1:],eabs[:-1])
           rats = concatenate([rats,rats1])

#    print
#    print sort(rats)
#    print
    qrats = quartiles(rats)
    print "Quartiles of ab ratios=",qrats
    efac = qrats[0]
#    efac = 1.23104806

    dups = {}
    newp = {}
    purge = {}

    for p in ws[:len(ws)]:
#    for p in []:
        if p in laps.keys():
           w2 = laps[p].keys()
           order = argsort([conall[w3].area for w3 in w2])
           w2 = [w2[j] for j in order] + [p]
           for j,w in enumerate(w2[:-1]):
#           w2 = [w2[j] for j in order]
#           for j,w in enumerate(w2):
               for v in w2[j+1:]:
                   vu = v
#                   if v in dups:
#                      vus = [v+"."+str(n+1) for n in range(dups[v])]
#                      vux = array([ellall[vu].mx for vu in vus])
#                      vuy = array([ellall[vu].my for vu in vus])
#                      vur = sqrt(vux**2+vuy**2)
#                      vu = vus[argmin(vur)]
#                   else:
#                      vu = v
                   interior = ellall[w].contour.multiply(efac).inside(conall[vu].x,conall[vu].y)
#                   if sum(interior) > 0.98*len(conall[vu].x):
#                      pass
                   if 12 < sum(interior) < len(conall[vu].x):
                      ix,iy = compress(interior, [conall[vu].x, conall[vu].y], 1)
                      iz = ones(iy.shape)
                      ellnew = Ellipse(ix,iy,iz,ns=0,verb=0)
                      areanew = pi*ellnew.a*ellnew.b
                      areaold = pi*ellall[w].a*ellall[w].b
#                      if areanew < areaold:
#                         purge.setdefault(w)
#                         print "here!!!!!"
#                         print v,w,sum(interior),len(interior),len(conall[vu].x)
#                         print ellall[v].mx,ellall[v].my,ellall[v].ab
#                         print ellall[w].mx,ellall[w].my,ellall[w].ab
#                         print ellnew.mx,ellnew.my,ellnew.ab
#                         print areanew,areaold
#                         print "here!!!!!"
                      if 1:
                         intnew = ellnew.contour.inside(conall[w].x,conall[w].y)
                         dx = (ellnew.mx-ellall[w].mx)
                         dy = (ellnew.my-ellall[w].my)
                         dr = sqrt(dx**2+dy**2)
                         if sum(interior) < len(conall[vu].x) and dr<0.1*ellall[w].ab:
                            interior = ellnew.contour.inside(ellall[w].ellx,ellall[w].elly)
                            if 0:
                               print
                               print "new ellipse!",vu,w,conall[w].mx,conall[w].my
                               print len(conall[vu].x),len(ix)
                               print ellall[vu].mx,ellall[vu].my,ellall[vu].ab
                               print ellnew.mx,ellnew.my,ellnew.ab
                               print sum(interior), len(interior),
                               print sum(1.*interior)/len(interior)
                            if sum(1.*interior) > 0.5*len(interior):
                                ellall[vu] = ellnew
                            else:
                                if v not in dups: dups[v] = 0
                                dups[v] += 1
                                v2 = v+"."+str(dups[v])
                                print "Adding",v2,dx,dy,dr
                                conall[v2] = copy.copy(conall[vu])
                                ellall[v2] = ellnew
                                if v2 not in ws:
                                   ws.append(v2)
                                if v2 not in laps:
                                   laps[v2] = {}
                                laps[v2].setdefault(w)
                                if w in laps[used[w]]:
                                   laps[used[w]].pop(w)
                                used[w] = v2

    for w in laps.keys():
        if not laps[w]: laps.pop(w)
    print "Number of families left",len(laps)

    print Set([len(laps[w]) for w in ws if w in laps])
    for w,n in dups.items():
        for k in range(n):
            w2 = w+"."+str(k+1)
            if w2 not in ws: ws.append(w2)

    print dups
    for w,n in dups.items():
        if w in laps:
           for k in range(n):
               w2 = w+"."+str(k+1)
               for v in laps[w2]:
                   if v in laps[w]:
                      laps[w].pop(v)
                      if not laps[w]: laps.pop(w)

    sp = {}
    [sp.__setitem__(l,sp.get(l,0)+1) for k in laps for l in laps[k]]
    print Set(sp.values())

#    for l in sp:
#        if sp[l] > 1:
#           print l,sp[l],ellall[l].mx,ellall[l].my,ellall[l].ab
#           for k in laps:
#               if l in laps[k]:
#                  print k,len(laps[k]),ellall[k].mx,ellall[k].my,ellall[k].ab

#    exit()
#    for w in purge:
#        ws.remove(w)
#        for k in laps:
#            if w in laps[k]: laps[k].pop(w)

    print "Prune families from",len(laps)
    popping = True
    pn = 0
    pruned = {}
    while popping:
       popping = False
       pn += 1
       if pn and not (pn%100):
          sys.stdout.write(".")
          sys.stdout.flush()
       for w in ws:
           if w in laps:
              for w2 in laps[w].keys():
                  if w!=w2 and w2 not in pruned and w2 in laps:
                     [laps[w].setdefault(w3) for w3 in laps[w2]]
                     [used.__setitem__(w3,w) for w3 in laps[w2]]
                     pruned.setdefault(w2)
                     laps.pop(w2)
                     popping = True
                     break
           if popping: break

    for w in laps.keys():
        if not laps[w]: laps.pop(w)
    print Set([len(laps[w]) for w in ws if w in laps])
    print "Number of families left",len(laps)

    sp = {}
    [sp.__setitem__(l,sp.get(l,0)+1) for k in laps for l in laps[k]]
    print Set(sp.values())
    print Set([len(laps[w]) for w in ws if w in laps])
#    for w in laps:
#        if len(laps[w]) == 0:
#           print ellall[w].mx, ellall[w].my, ellall[w].a, ellall[w].b, ellall[w].theta
#    exit()

    inconseq = {}
    print "Refit ellipses?"
    if 0:
       for w in ws:
           if w in laps:
              interior = zeros(conall[w].x.shape,float)
#              print w,conall[w].mx,conall[w].my,conall[w].r,ellall[w].ab,conall[w].area
              for v in laps[w]:
                  dr = sqrt((conall[v].mx-conall[w].mx)**2+(conall[v].my-conall[w].my)**2)
#                  print v,conall[v].mx,conall[v].my,conall[v].r,ellall[v].ab,dr,
                  if (2*conall[v].r+dr) > 0.5*conall[w].r and conall[v].r < 0.25*conall[w].r:
#                  if dr > conall[w].r:
                     interior1 = conall[v].inside(conall[w].x,conall[w].y)
#                     print w,v,sum(interior1),
                     interior1 = conall[v].multiply(2).inside(conall[w].x,conall[w].y)
#                     print sum(interior1),
                     interior = interior + interior1
#                     print sum(interior),
#                  print
              ix,iy = compress(equal(interior,0), [conall[w].x, conall[w].y], 1)
#              print w,len(conall[w].x),len(ix),
              if not len(ix):
                 print "kill!",conall[w].mx,conall[w].my
                 inconseq.setdefault(w)
                 if w in laps:
                    [inconseq.setdefault(w3) for w3 in laps[w]]
              elif len(ix)!=len(conall[w].x):
                 iz = ones(iy.shape)
                 ellall[w] = Ellipse(ix,iy,iz,ns=0,verb=0)
#              print

    ws = [w for w in ws if w not in inconseq]

    for w in ws:
        xpixels, ypixels, zpixels = ellall[w].contour.pixels(data)
        wpixels = divz(zpixels*greater(zpixels,0),sqrt((getsqrt(zpixels)**2+sigma**2)))
        upixels = greater(wpixels,0)
        xpixels,ypixels,wpixels,zpixels = compress(upixels, \
                [xpixels,ypixels,wpixels,zpixels], 1)
        ellall[w].counts = sum(zpixels)

    cleaned = {}
    inconseq = {}
    watchout = {}

    for j,w in enumerate(ws):
        cleaning = True
#        cleaning = False
        while cleaning:
           cleaning = False
           if w in laps and len(laps[w]):
              xs,ys,fs = 0.0,0.0,0.0
              xl,yl,fl = [],[],[]
              for v in [w]+laps[w].keys():
                  dr = sqrt((ellall[v].mx-ellall[w].mx)**2+(ellall[v].my-ellall[w].my)**2)
                  f = sqrt(ellall[v].v*ellall[v].ab)*exp(-0.5*dr**2/ellall[w].ab**2)
                  fs += f
                  xs += ellall[v].mx*f
                  ys += ellall[v].my*f
                  xl.append(ellall[v].mx)
                  yl.append(ellall[v].my)
                  fl.append(f)
              xc,yc = divz(xs,fs), divz(ys,fs)
              xl,yl,fl = asarray([xl,yl,fl])
#              xc = median(xl)
#              yc = median(yl)
              xs,ys,fs = 0.0,0.0,0.0
              xl,yl,fl = [],[],[]
              for v in [w]+laps[w].keys():
                  dr = sqrt((ellall[v].mx-ellall[w].mx)**2+(ellall[v].my-ellall[w].my)**2)
                  f = (ellall[v].v*ellall[v].ab)*exp(-0.5*dr**2/ellall[w].ab**2)
                  fs += f
                  xs += (ellall[v].mx-xc)**2*f
                  ys += (ellall[v].my-yc)**2*f
                  xl.append(ellall[v].mx-xc)
                  yl.append(ellall[v].my-yc)
                  fl.append(f)
              xr,yr,rr = sqrt(divz(xs,fs)), sqrt(divz(ys,fs)), sqrt(divz(xs+ys,fs))
              xl,yl,fl = asarray([xl,yl,fl])
#              xr = 1.49*median(abs(xl))
#              yr = 1.49*median(abs(yl))
#              rr = 1.49*median(xl**2+yl**2)
              print "%20s %4d %10.3f %9.3f %10.3f %9.3f %9.3f" % (w,len(laps[w]),xc,xr,yc,yr,rr),
              bad = 0
              purge = {}
              purged = {}
              sf = 0.0
              killoff = False
              wl = laps[w].keys()
              wl = [wl[j2] for j2 in argsort([ellall[w1].ab for w1 in wl])[::-1]]
              for v in wl:
                  dx = (ellall[v].mx - xc)
                  dy = (ellall[v].my - yc)
                  dr = sqrt(dx**2+dy**2)/max([ellall[v].ab,2.0])
#                  df = divz(ellall[v].counts,cint)
#                  dc = divz(ellall[v].counts,divz(cint,care))
#                  nokeep = (dr>1.0) and (df>0.25 or (dc and dc>1.0))
                  nokeep = (dr>1.0)
                  if nokeep:
#                     if df > 1.00: killoff = True
                     purge.setdefault(v)
                     bad += 1
                     cleaning = True
              cint = 0.0
              care = 0.0
              wl = [v for v in wl if v not in purge]
              for v in wl:
                  dx = (ellall[v].mx - xc)
                  dy = (ellall[v].my - yc)
                  if sqrt(dx**2+dy**2)<=max([2*rr,4]):
                      cint = max([0.0,ellall[w].counts - ellall[v].counts])
                      care = max([0.0,ellall[w].area - ellall[v].area])
                      break
              print cint,
#              for v in purge:
#                  print v,ellall[v].mx, ellall[v].my, ellall[v].ab
              if cint <= 0.0: killoff = True
#              purge2 = {}
#              laps2 = {}
#              [laps2.setdefault(w3) for w2 in laps for w3 in laps[w2] if w2 != w]
#              [purge2.setdefault(v) for v in purge if v in laps2]
#              pcf = lambda v: 1.*sum(ellall[w].contour.inside(ellall[v].x,ellall[v].y))/len(ellall[v].x)
#              print [pcf(v) for v in purge2],
#              pc = sum([pcf(v)*ellall[v].counts for v in purge2])/ellall[w].counts
#              if pc > 4.00: killoff = True
#              print pc
#              for v in purge2:
#                  print v,ellall[v].mx, ellall[v].my, ellall[v].ab, ellall[v].counts
#              print
              while purge:
                 vs = purge.keys()
                 vs = [vs[j2] for j2 in argsort([ellall[v].area for v in vs])[::-1]]
                 if len(vs) > 1 and vs[0] not in laps:
                    laps[vs[0]] = {}
                    for v in vs[1:]:
                       laps[vs[0]].setdefault(v)
                    cleaning = True
#                     print conall[v].mx,conall[v].my,conall[v].r
#                     if purge[v] and v not in laps:
#                        killoff = True
#                        lap1 = GetOverlaps(w,ws[j+1:],conall,ellall)
#                        laps.setdefault(v,lap1)
#                        [inconseq.setdefault(v2) for v2 in laps if v in laps[v2]]
                 for v in vs:
                     laps[w].pop(v)
                     purge.pop(v)
                     purged.setdefault(v)
                     if v in laps:
                        for v2 in laps[v]:
                            if v2 in laps[w]:
#                               print conall[v2].mx,conall[v2].my,conall[v2].r
                               purge.setdefault(v2)
              if purged:
                 for v in [w]+laps[w].keys():
                     if v not in watchout: watchout[v] = {}
                     for v2 in purged:
                         watchout[v].setdefault(v2)
#                 for w2 in [w]+laps[w].keys():
#                 for w2 in []:
#                     dx = (conall[w2].mx - xc)
#                     dy = (conall[w2].my - yc)
#                     if w2 not in purged and sqrt(dx**2+dy**2)<=max([2*rr,4]):
#                        interior = zeros(conall[w2].x.shape,float)
#                        print
#                        for v in purged:
#                            interior1 = conall[v].inside(conall[w2].x,conall[w2].y)
#                            print w,v,sum(interior1),
#                            interior1 = conall[v].multiply(2).inside(conall[w2].x,conall[w2].y)
#                            print sum(interior1),
#                            interior = interior + interior1
#                            print sum(interior)
#                        ix,iy = compress(equal(interior,0), [conall[w2].x, conall[w2].y], 1)
#                        print len(conall[w2].x),len(ix)
#                        if len(ix)!=len(conall[w2].x):
#                           iz = ones(iy.shape)
#                           ellall[w2] = Ellipse(ix,iy,iz,ns=0,verb=0)
#                        elif not len(ix):
#                           inconseq.setdefault(w2)
#                           if w2 in laps:
#                              [inconseq.setdefault(w3) for w3 in laps[w2]]
              if killoff:
                 inconseq.setdefault(w)
                 [inconseq.setdefault(w2) for w2 in laps[w]]
              if bad: cleaned[w] = True
              print bad

    print "Refit ellipses?"
    if 0:
       for w in watchout:
           interior = zeros(conall[w].x.shape,float)
           for v in watchout[w]:
               interior1 = conall[v].inside(conall[w].x,conall[w].y)
#               print w,v,sum(interior1),
               interior1 = conall[v].multiply(2).inside(conall[w].x,conall[w].y)
#               print sum(interior1),
               interior = interior + interior1
#               print sum(interior)
           ix,iy = compress(equal(interior,0), [conall[w].x, conall[w].y], 1)
           print w,len(conall[w].x),len(ix),
           if not len(ix):
              print "kill!",conall[w].mx,conall[w].my
              inconseq.setdefault(w)
              if w in laps:
                 [inconseq.setdefault(w3) for w3 in laps[w]]
           elif len(ix)!=len(conall[w].x):
              iz = ones(iy.shape)
              ellall[w] = Ellipse(ix,iy,iz,ns=0,verb=0)
           print

    print "Number cleaned=",len(cleaned)
    print "Number inconseq=",len(inconseq)

    [laps.pop(w) for w in inconseq if w in laps]

    print
    print len(laps)
    mabs = {}
    for j,w in enumerate(ws):
        if w in laps and len(laps[w]) > 2:
           wabs = array([ellall[k].ab for k in laps[w]])
           wvs = array([ellall[k].v for k in laps[w]])
           wabs,wvs = take([wabs,wvs],argsort(wabs),1)
           warea = pi*wabs**2
           wf = add.accumulate( warea*wvs)
           wr = sqrt(divz(sum(wf*wabs**2),sum(wf)))
           if wr > 0: mabs[w] = wr

    wrs = sort(array([mabs[w] for w in mabs]))
    print "wrs"
    print "quartiles(wrs)",quartiles(wrs)
    print "extrema(wrs)",extrema(wrs)
    ars0 = arange(-1.0,2.0,0.05,float32)
    nrs = diff(1.0*searchsorted(log10(wrs),ars0))
    ars = diff(ars0)/2.0 + ars0[:-1]
    crs = add.accumulate(nrs)/sum(nrs)
    nrs,ars,crs = compress(nrs,[nrs,ars,crs],1)

    if less(diff(nrs),0).any():
       see = 10.0**ars[nonzero(less(diff(nrs),0))[0]]
       print "Radius=",see

    print "Prune families from",len(laps)
    popping = True
    pn = 0
    pruned = {}
    while popping:
       popping = False
       pn += 1
       if pn and not (pn%100):
          sys.stdout.write(".")
          sys.stdout.flush()
       for w in ws:
           if w in laps:
              for w2 in laps[w].keys():
                  if w!=w2 and w2 not in pruned and w2 in laps:
                     [laps[w].setdefault(w3) for w3 in laps[w2]]
                     [used.__setitem__(w3,w) for w3 in laps[w2]]
                     pruned.setdefault(w2)
                     laps.pop(w2)
                     popping = True
                     break
           if popping: break

    print "Number of families left",len(laps)

    parents = {}

    nl = len(laps)
    vn = 10.0**(int(log10(nl+1)+1))
    n = len(ws)
    value = 0
    used = {}
    drawings = []
    colors = []
    fills = []
    for j in range(n):
        if ws[j] not in used and ws[j] not in inconseq:
           uw = [ws[j]]
           luw,nuw = 0,1
           while nuw!=luw:
              luw = len(uw)
              for w2 in uw:
                 if w2 in laps:
                    [uw.append(w) for w in laps[w2] if w not in uw]
              nuw = len(uw)
           uw = [w for w in uw if w not in used]
           if uw:
              value = value+1
#              print value,len(uw),len(used)
#              sys.stdout.flush()
              order = argsort([conall[w].area for w in uw])[::-1]
              uw = [uw[j2] for j2 in order]
              parents[uw[0]] = {}
              for w in uw:
                  parents[uw[0]][w] = (conall[w],ellall[w])
                  used.setdefault(w)
              uw = [w for w in uw if ellall[w].ab > 1]
              for w in uw:
                  if w == uw[0] or w == uw[-1]:
                     if efit:
                        color = conall[w].v
                        x, y = ellall[w].ellx, ellall[w].elly
                        x = clip(x,0,naxis1-1)
                        y = clip(y,0,naxis2-1)
#                        print w,extrema(x), extrema(y)
                        color = value/vn
                        color = conall[w].v
                        color = -999999
                        segments = asarray([[(x[j],x[j+1]),(y[j],y[j+1])] for j in range(len(x)-1)])
                     else:
                        color = value/vn
#                        color = conall[w].v
                        color = -999999
                        segments = conall[w].contour
#                        segments = conall[w].multiply(2.0).contour
                     drawings.append(segments)
                     colors.append(color)
                     if w == uw[-1]:
                        if efit:
#                           fill = value/vn
#                           fill = conall[w].v
                           fill = ellall[w].counts/ellall[w].area
                        else:
                           fill = ellall[w].counts/ellall[w].area
                     else:
                         fill = 0 
                     fills.append(fill)

    oframe = zeros((naxis2,naxis1),float64)
    if len(drawings):
       print
       print len(drawings)
       sys.stdout.flush()
       oframe = VTKDrawContour(zeros(oframe.shape,float64),drawings,color=colors,fill=fills,verb=1).astype(float32)
    print
    sys.stdout.flush()
    if efit: oimage = "efamilies%04d.fits" % (n1)
    else: oimage = "families%04d.fits" % (n1)
    qdump(oimage,oframe)
    zdb = "families%04d.zodb" % (n1)
    if efit: WriteContours(zdb,parents)

def ModelFamilies(image,zdb,catalog,dsize,ext0,magmax,abmin,bamin,verb=0):
    print "Reading",image
    mimage = image.replace(".fits","models.fits")
    rimage = image.replace(".fits","resids.fits")
    F = FITS(image)
    naxis1 = F["naxis1"]
    naxis2 = F["naxis2"]
    data = F.data()
    filt = F["filter"]
    if filt == "N/A" or "CLEAR" in filt:
       filt = F["filter1"]
    if filt == "N/A" or "CLEAR" in filt:
       filt = F["filter2"]
    filt = filt.strip()
    F.close()

    catalog = readtable(catalog,header="# id",tcols=["id","ra","dec"])
    sizes = {}
#    [sizes.setdefault(o,(ab,ba,mag)) for o,ab,ba,mag in zip( catalog["id"],catalog["mr"],catalog["mb"]/catalog["ma"],catalog["mag"])]
    [sizes.setdefault(o,(ab,ba,mag)) for o,ab,ba,mag in zip( catalog["id"],catalog["ab"],catalog["ba"],catalog["mag"])]
    stars = {}
    [stars.setdefault(o) for o,s in zip( catalog["id"],catalog["star"]) if s]
    galaxies = {}
    [galaxies.setdefault(o) for o,s in zip( catalog["id"],catalog["star"]) if not s]
    print "Number of stars=",len(stars)

    models = zeros(data.shape,float64)
    resids = zeros(data.shape,float64)

    parents,conn,db,storage = ReadContours(zdb,verb=0,noclose=True)
    print

    ws = parents.keys()
    mao = []
    for w in ws:
        uw = parents[w].keys()
        mys = [parents[w][v][1].my for v in uw]
        ars = [parents[w][v][1].area for v in uw]
        mao.append( max(ars))
    order = argsort(mao)[::-1]
    ws = [ws[j] for j in order]
    print "Have %d families" % (len(ws))
    ws = [w for w in ws if w in sizes]
    print "Have %d families" % (len(ws))
    ws = [w for w in ws if sizes[w][0] > 1.0]
    print "Have %d families" % (len(ws))
    gws = [w for w in ws if w in galaxies]
    print "Have %d families" % (len(gws))
    mws = [w for w in gws if sizes[w][2] <= magmax]
    print "Have %d families" % (len(mws))
    mws = [w for w in mws if sizes[w][0] >= abmin]
    print "Have %d families" % (len(mws))
    mws = [w for w in mws if sizes[w][1] >= bamin]
    print "Have %d families" % (len(mws))

    for w in mws:
        children = parents[w]
        parent = children[w]
        px,py = parent[1].mx, parent[1].my
        pab = parent[1].ab
        pab = sizes[w][0]
        ext = ext0*(1+divz(2*pab,dsize)**2)
#        ext = ext0
#        if pab > dsize:
#           ext = 2*ext
#        mag = ext*floor(power(pab/16.0+1,1))
        mag = ext
        p2 = parent[1].contour.multiply(mag)
        mw = max([diff(p2.ex)[0], diff(p2.ey)[0]])
        print w,px,py,pab,ext,mag,mw
        vmask = []
        ai = 0.0
        for v in ws:
            if v not in [w]+children.keys():
               c = parents[v][v][1]
               bx = between(c.contour.ex,p2.ex[0],p2.ex[1])
               by = between(c.contour.ey,p2.ey[0],p2.ey[1])
               if c.ab > 2 and bx.any() and by.any():
                  cm = 2*sizes[v][0]/parents[v][v][1].ab
                  c2 = parents[v][v][1].contour.multiply(cm)
#                  c2 = c.contour.multiply(2)
                  i = p2.inside(c2.x,c2.y)
                  if i.any():
                     print v,c.mx,c.my,c.ab
                     ai += pi*(cm*c.ab)**2 * sum(i)/len(c2.x)
                     vmask.append(v)
        ap = pi*(mag*pab)**2
        print
        print ai,ap,ai/ap
        print
        if (ai/ap)<0.5:
           ll = floor([p2.ex[0],p2.ey[0]])
           ur = floor([p2.ex[1]+1,p2.ey[1]+1])
           subsec = 1*data[p2.ey[0]:p2.ey[1],p2.ex[0]:p2.ex[1]]
           children = parents[w].keys()
           yi,xi = indices(subsec.shape)
           xi = xi + p2.ex[0] + 0
           yi = yi + p2.ey[0] + 0
           maskim = zeros(subsec.shape,float64)
           print "len(vmask)=",len(vmask)
           for v in vmask:
               sys.stdout.write(".")
               sys.stdout.flush()
               c = parents[v][v][1].contour
               cm = sizes[v][0]/parents[v][v][1].ab
#               cm = 2.0*cm
               c2 = parents[v][v][1].contour.multiply(cm)
               c2x0,c2x1 = clip(floor(add(c2.ex-p2.ex[0]-0,[0,1])).astype(int32),0,xi.shape[1]-1)
               c2y0,c2y1 = clip(floor(add(c2.ey-p2.ey[0]-0,[0,1])).astype(int32),0,yi.shape[0]-1)
               sxi = xi[c2y0:c2y1+1,c2x0:c2x1+1]
               syi = yi[c2y0:c2y1+1,c2x0:c2x1+1]
               si = asarray([c2.inside(xi1,yi1) for xi1,yi1 in zip(sxi,syi)])
               i = zeros(maskim.shape,float64)
               i[c2y0:c2y1+1,c2x0:c2x1+1] = si
               add(maskim,i,maskim)
           print
           if verb:
              qdump("/tmp/subsec.fits",1.*subsec)
              qdump("/tmp/maskim.fits",1.*maskim)
           maskim = equal(maskim,0)*not_equal(subsec,0)
           g = GalPhot()
           g.initparams()
           cx = array([parents[w][v][1].mx for v in children])
           cy = array([parents[w][v][1].my for v in children])
           cr = array([parents[w][v][1].ab for v in children])
           g.xcen = cx[argmin(cr)] - p2.ex[0] + 1
           g.ycen = cy[argmin(cr)] - p2.ey[0] + 1
           print maskim.shape,mw,sum(maskim.ravel()),len(maskim.ravel())
           g.r1 = 1.0*pab
           g.r2 = 0.0
           g.rmin = 0.3
           g.rmax = ext*pab
           g.hmax = 4
           g.cliplow = 0.0
           g.cliphigh = 0.0
           g.rshap = 0.75*ext*pab
           g.rcen = 0.75*ext*pab
           g.ellipsefit = 1
           g.image = where(maskim,subsec,VALBAD).astype(float32)
           gt0 = time.time()
           g.photfit()
           gt1 = time.time()
           print "GALPHOT Time = %.3f" % (gt1-gt0)
           if verb:
              qdump("/tmp/resim.fits",g.resim+g.i[-1,0])
           h = g
           h.inputparams(g)
           h.image = subsec
           h.justres = 1
           ht0 = time.time()
           h.photfit()
           ht1 = time.time()
           print "GALPHOT Time = %.3f" % (ht1-ht0)
           model1 = subsec-(h.resim+h.i[-1,0])
           data[p2.ey[0]:p2.ey[1],p2.ex[0]:p2.ex[1]] = \
              data[p2.ey[0]:p2.ey[1],p2.ex[0]:p2.ex[1]] - model1
           if verb:
              qdump("/tmp/resim2.fits",h.resim+h.i[-1,0])
              qdump("/tmp/model1.fits",model1)
           models[p2.ey[0]:p2.ey[1],p2.ex[0]:p2.ex[1]] = \
               models[p2.ey[0]:p2.ey[1],p2.ex[0]:p2.ex[1]] + model1
           if verb:
              print "Writing",mimage
              F = FITS(mimage,create=1,template=image)
              F.putdata(models)
              F.close()
              print "Writing",rimage
              F = FITS(rimage,create=1,template=image)
              F.putdata(data)
              F.close()
           if verb:
#              exit()
              raw_input("<hit cr> ")

    print "Writing",mimage
    F = FITS(mimage,create=1,template=image)
    F.putdata(models)
    F.close()

    print "Writing",rimage
    F = FITS(rimage,create=1,template=image)
    F.putdata(data)
    F.close()


def GetOverlaps(w,ws,conall,ellall,e=0):
    con = conall[w]
    ell = ellall[w]
    if e: (wxb,wyb),(wxe,wye) = transpose(con.contour)
    else: (wxb,wyb),(wxe,wye) = transpose(ell.contour.contour)
    wx, wy = wxb,wyb
    ewx,ewy = extrema(wxe), extrema(wye)
    laps = {}
    check = False
    for v in ws:
        if e: vx,vy = ellall[v].mx, ellall[v].my
        else: vx,vy = conall[v].mx, conall[v].my
        if v == w:
           box = True
        else:
#           box = between(conall[v].x,ewx[0],ewx[1]) * between(conall[v].y,ewy[0],ewy[1])
#           box = box.any()
           if e:
             box = (ellall[v].contour.ex[1] >= ell.contour.ex[0] and ellall[v].contour.ex[0] <= ell.contour.ex[1]) \
               and (ellall[v].contour.ey[1] >= ell.contour.ey[0] and ellall[v].contour.ey[0] <= ell.contour.ey[1])
           else:
             box = (conall[v].ex[1] >= con.ex[0] and conall[v].ex[0] <= con.ex[1]) \
               and (conall[v].ey[1] >= con.ey[0] and conall[v].ey[0] <= con.ey[1])
#        if v==w or (conall[v].ex[1] >= con.ex[0] and conall[v].ex[0] <= con.ex[0] and \
#           conall[v].ey[1] >= con.ey[0] and conall[v].ey[0] <= con.ey[0]):
#        if ewx[0]<=vx<=ewx[1] and ewy[0]<=vy<=ewy[1]:
        if box:
           if v!=w:
              check = True
              va = sum(mangle(arctan2(wye-vy,wxe-vx) - arctan2(wyb-vy,wxb-vx)))
              interior = False
              if abs(va-2*pi)<1e-8:
                 interior = True
              else:
                 if e: xb,yb = ellall[v].contour.x,ellall[v].contour.y
                 else: xb,yb = conall[v].x,conall[v].y
#                 t = Interiors(xb,yb,con.contour,verb=0)
#                 interior = less(abs(t-2*pi),1e-4).any()
                 interior = con.inside(xb,yb).any()
#              print w,v
#              print conall[v].mx, conall[v].my, con.mx, con.my, con.wid, con.hei
#              print interior
#              print va
           if v == w or interior:
              laps.setdefault(v)
#        if ewx[0]<=vx<=ewx[1] and ewy[0]<=vy<=ewy[1]:
##           va = sum(mangle(arctan2(wy[1:]-vy,wx[1:]-vx) - arctan2(wy[:-1]-vy,wx[:-1]-vx)))
#           va = sum(mangle(arctan2(wye-vy,wxe-vx) - arctan2(wyb-vy,wxb-vx)))
#           if abs(va-2*pi)<1e-8:
#              laps.setdefault(v)
#    if check:
#       print laps
#       raw_input()
    return laps

def NestOverlaps(laps,c):
    d = 0
    print c
    for w in laps:
        if laps[w]:
           for v in laps[w]:
               if v in laps and laps[w][v] != laps[v]:
                  laps[w][v] = laps[v]
                  d = 1
    if d: c = c+1
    else: c = 0
    return laps,c


def AddContours( image,n1,n0,levels):
    F = FITS(image)
    naxis1 = F["naxis1"]
    naxis2 = F["naxis2"]
    sigma = F["bnoise"]
    F.close()
    print "Getting contours and ellipses for",image
    sys.stdout.flush()
    contours, ellipses = GetContours( image,n1,n0,levels)
    print "Got contours and ellipses for",image
    sys.stdout.flush()

    oframe = zeros((naxis2,naxis1),float64)
    print "Noise=",sigma
    for level in levels:
        drawings = []
        for j in contours[level]:
            drawings.append(contours[level][j].contour)
            if not ellipses[level][j].fail:
               x = ellipses[level][j].ellx
               y = ellipses[level][j].elly
               segments = asarray([[(x[j],x[j+1]),(y[j],y[j+1])] for j in range(len(x)-1)])
#               drawings.append(segments)
        if len(drawings):
           print "Drawing",len(drawings)
           eframe = VTKDrawContour(zeros(oframe.shape,float64),drawings,color=level*sigma,fill=0,verb=1).astype(float32)
           add(oframe,eframe,oframe)
    oimage = "ellipses%04d.fits" % (n1)
    qdump(oimage,oframe)

def FitEllipse(Xt,Yt,Wt):
    f = sum(Wt)
    mx = sum(Wt*Xt)/f
    my = sum(Wt*Yt)/f
    sx = sqrt(2*sum(power(Xt-mx,2)*Wt)/f)
    sy = sqrt(2*sum(power(Yt-my,2)*Wt)/f)
    if not sx: sx = 1
    if not sy: sy = 1
    Xr = (Xt-mx)/sx
    Yr = (Yt-my)/sy
    D = transpose([Xr*Xr, Xr*Yr, Yr*Yr, Xr, Yr, ones(Xr.shape)])
#    mm = lambda a1,a2: matrixmultiply(a1,a2)
    mm = lambda a1,a2: dot(a1,a2)
    D = D*Wt[::,NewAxis]
    Dt = transpose(D)
    S = mm(Dt,D)
    C = zeros((6,6),float64)
    C[0,2] = -2.0
    C[1,1] = 1.0
    C[2,0] = -2.0
    tmpA = S[:3,:3]
    tmpB = S[:3,3:]
    tmpC = S[3:,3:]
    tmpD = C[:3,:3]
    if not determinant(tmpC): return -99,-99,-99,-99,-99,-99,-99
    tmpE = mm(inverse(tmpC),transpose(tmpB))
    eigval,eigvec = eigenvectors(mm(inverse(tmpD),(tmpA-mm(tmpB,tmpE))))
    neg = nonzero(less(eigval.real,0))
    if not len(neg): return -99,-99,-99,-99,-99,-99,-99
    ie = neg[0]
    A = eigvec[ie].real
    A = concatenate([A,mm(-tmpE,A)])
    par = asarray([ A[0]*sy*sy,
          A[1]*sx*sy,
          A[2]*sx*sx,
          -2*A[0]*sy*sy*mx - A[1]*sx*sy*my + A[3]*sx*sy*sy,
          -A[1]*sx*sy*mx - 2*A[2]*sx*sx*my + A[4]*sx*sx*sy,
          A[0]*sy*sy*mx*mx + A[1]*sx*sy*mx*my + A[2]*sx*sx*my*my
          - A[3]*sx*sy*sy*mx - A[4]*sx*sx*sy*my
          + A[5]*sx*sx*sy*sy ])
    theta = 0.5*arctan2(par[1],par[0]-par[2])
    cost = cos(theta); sint = sin(theta); sint2 = sint**2; cost2 = cost**2; sincost = sint*cost
    Ao = par[5]
    Au = par[3]*cost + par[4]*sint
    Av =-par[3]*sint + par[4]*cost
    Auu = par[0]*cost2 + par[2]*sint2 + par[1]*sincost
    Avv = par[0]*sint2 + par[2]*cost2 - par[1]*sincost
    if not Auu or not Avv: return -99,-99,-99,-99,-99,-99,-99
    tu = -Au/(2*Auu)
    tv = -Av/(2*Avv)
    wc = Ao - Auu*tu**2 - Avv*tv**2
    u = tu*cost - tv*sint
    v = tu*sint + tv*cost
    Ru = -wc/Auu
    Rv = -wc/Avv
    Ru = sqrt(abs(Ru))*sign(Ru)
    Rv = sqrt(abs(Rv))*sign(Rv)
    mx,my,f,r,theta,a,b = u,v,f,sqrt(Ru*Rv),theta*180/pi,Ru,Rv
    if b > a: a,b,theta = b,a,theta+90.0
    return mx,my,f,r,theta,a,b

class Contour(persistent.Persistent):
    def __init__(self,contour,v):
        self.contour = contour
        self.v = v
        (self.xb,self.yb),(self.xe,self.ye) = transpose(contour)
        self.x = concatenate([self.xb,self.xb[:1]])
        self.y = concatenate([self.yb,self.yb[:1]])
#        self.getcenter()
#        self.mx,self.my = mean(self.x),mean(self.y)
        self.mx,self.my = median(self.x),median(self.y)
        self.r = sqrt(sum((self.x-self.mx)**2 + (self.y-self.my)**2)/len(self.x))
        self.ex = extrema(self.x)
        self.ey = extrema(self.y)
        self.wid = diff(self.ex)[0]
        self.hei = diff(self.ey)[0]
        self.box = self.wid*self.hei
        x1,y1 = transpose(asarray(contour)[::,::,0])
        x2,y2 = transpose(asarray(contour)[::,::,1])
        self.area = 0.5*abs(sum(x1*y2-y1*x2))
    def getcenter(self):
        segments = self.contour
        for j in range(1):
            (xb,yb),(xe,ye) = transpose(segments)
            x = 0.5*(xb+xe)
            y = 0.5*(yb+ye)
            mx,my = median(x),median(y)
#            print extrema(x),mx,extrema(y),my
            x = concatenate([x,x[:1]])
            y = concatenate([y,y[:1]])
            segments = asarray([[(x[j],x[j+1]),(y[j],y[j+1])] for j in range(len(x)-1)])
        self.mx,self.my = mx,my
    def multiply(self,fac):
        (xb,yb),(xe,ye) = transpose(self.contour)
        x = concatenate([xb,xb[:1]])
        y = concatenate([yb,yb[:1]])
        mx,my = self.mx,self.my
        xb,xe = fac*(xb-mx)+mx, fac*(xe-mx)+mx
        yb,ye = fac*(yb-my)+my, fac*(ye-my)+my
        contour = transpose([[xb,yb],[xe,ye]])
        return Contour(contour,self.v)
    def inside(self,xa,ya):
        t = Interiors(xa,ya,self.contour,verb=0)
        interior = less(abs(t-2*pi),1e-4)
        return interior
    def pixels(self,data):
        x0,x1 = int(self.ex[0]), int(self.ex[1]+1)
        y0,y1 = int(self.ey[0]), int(self.ey[1]+1)
        xpixels, ypixels, zpixels = [],[],[]
        for y in arange(y0,y1,1,int32):
            ua = between(self.y,y-2,y+2,inc=1)
            if ua.any():
               x0,x1 = extrema(compress(ua,self.x))
               x0,x1 = clip([int(x0-1),int(x1+2)],0,data.shape[1])
               xa = arange(x0,x1,1,float32)
               ya = y*ones(xa.shape)
               interior = self.inside(xa,ya)
               if interior.any():
                  za = data[y,x0:x1]
                  ya,xa,za = compress(interior,[ya,xa,za],1)
                  ypixels = concatenate([ypixels,ya])
                  xpixels = concatenate([xpixels,xa])
                  zpixels = concatenate([zpixels,za])
        return asarray([xpixels, ypixels, zpixels])


class Ellipse(persistent.Persistent):
    def __init__(self,x,y,w,verb=0,fit=1,model=1,dth=2.0,ns=0):
        self.x = x
        self.y = y
        self.w = w
        self.mx = mean(x)
        self.my = mean(y)
        self.ns = ns
        self.v = mean(self.w)
        self.dth = dth
        self.verb = verb
#        if 2510 < self.mx < 2550 and 2530 < self.my < 2560:
#           self.verb  = 1
        if fit: self.fit()
        if model and self.mx!=99:
           self.contour = self.model()
        else: self.fail = True
    def __repr__(self):
        if hasattr(self,"ellipse"):
            fmt = " ".join(7*["%10.3f"])
            return fmt % tuple(self.ellipse)
        else: return "ellipse"
    def fit(self):
        ellipse = FitEllipse(self.x,self.y,self.w)
        self.ellipse = ellipse
        fmt = " ".join(7*["%10.3f"])
        if self.verb:
           print fmt % tuple(ellipse)
        if self.ns:
           h = min([len(self.x)/2,25])
           h = 15
           ellipses = []
           for j in range(self.ns):
               x,y,w = take([self.x,self.y,self.w],randint(0,len(self.x),(h,)),1)
               ellipse = FitEllipse(x,y,w)
               ellipses.append(ellipse)
           ellipses = asarray(ellipses)
           use = argsort(ellipses[::,3])[:3]
           self.ellipse = mean(take(ellipses,use,0))
           self.rms = 1.49*median(abs(ellipses-self.ellipse[NewAxis,::]))
           if self.verb:
              print fmt % tuple(self.ellipse)
              print fmt % tuple(self.rms)
              print
        self.mx,self.my,self.f,self.r,self.theta,self.a,self.b = self.ellipse
        self.ab = sqrt(self.a*self.b)
        self.area = pi*self.a*self.b
        self.cth = cos(self.theta*pi/180)
        self.sth = sin(self.theta*pi/180)
        if self.mx == -99 or self.a <= 0 or self.b <= 0 or self.ab <= 0:
           self.fail=True
           self.ab = -99
        else:
           self.fail = False
    def model(self):
        th = arange(0.0,360.0,self.dth,float)
        if th[-1] != 360.0: th = concatenate([th,[360.0]])
        cth = cos(th*pi/180)
        sth = sin(th*pi/180)
        x,y = self.a*cth, self.b*sth 
        self.ellx,self.elly = x*self.cth - y*self.sth + self.mx, \
                x*self.sth + y*self.cth + self.my
        xb,xe = self.ellx[:-1], self.ellx[1:]
        yb,ye = self.elly[:-1], self.elly[1:]
        contour = transpose([[xb,yb],[xe,ye]])
        return Contour(contour,self.v)

