#!/usr/bin/env python

### Tools to generate and query
### tables made with PyTables

### Currently catered for 2MASS PSC files

### Written by Edward Villanueva
### Last updated 01-18-11

from tables import *
from glob import glob
from numpy import asarray

import gzip
import os,sys
import time

t0 = time.time()
ifile = None
ofile = None
par1 = None
par2 = None
prange1 = None
prange2 = None
convert = None
hname = None
query = 0
new = 0
lp = 0
pr = 0
force = 0
tbl = None
all = 0
mkdict = 0

base = os.getenv("TWODIR")
if not base: 
   base = "/Volumes/Four/edwardv/2MASS"
ddir = base+"/downloaded"
colfile = base+"/psc_cols.txt"

def printtime():
    t1 = time.time()
    print "Elapsed time: %.2f seconds" % (t1-t0)

def readpsc(f):
    print "Reading in %s" % (f)
    if ".gz" in f:
       g = gzip.open(f,'rb')
       gl = g.readlines()
       g.close()
    else:
       f = open(f,"r")
       gl = f.readlines()
       f.close()
    print "Read in %s" % (f)
    return gl

def readcolnames():
    f = open(colfile,'r')
    fl = f.readlines()
    f.close()
    cols = [l.split()[0] for l in fl]
    return cols

def gettype(itype):
    scol = lambda x: "StringCol(%d)" % (x)
    T = { "real":"Float32Col()", \
	  "double precision":"Float64Col()", \
	  "smallint":"Int16Col()", \
	  "integer":"Int64Col()", \
	  "date":"StringCol(10)" \
	}
    if "character" in itype:
       length = int(itype.split("(")[1].rstrip(")"))
       return scol(length)
    else:
       return T[itype]

def gettype2(col):
    if col in ["ra","decl","err_maj","err_min","j_m","j_cmsig","j_msigcom","j_snr", \
    "h_m","h_cmsig","h_msigcom","h_snr","k_m","k_cmsig","k_msigcom","k_snr", \
    "prox","glon","glat","x_scan","jdate","j_psfchi","h_psfchi","k_psfchi", \
    "j_m_stdap","j_msig_stdap","h_m_stdap","h_msig_stdap","k_m_stdap", \
    "k_msig_stdap","dist_opt","b_m_opt","vr_m_opt"]: ctype = float

    elif col in ["err_ang","pxpa","pxcntr","gal_contam","mp_flg","pts_key","scan", \
    "dist_edge_ns","dist_edge_ew","dup_src","use_src","phi_opt", \
    "nopt_mchs","ext_key","scan_key","coadd_key","coadd"]: ctype = int
    
    elif col in ["designation","ph_qual","rd_flg","bl_flg","cc_flg","ndet","hemis", \
    "date","dist_edge_flg","a"]: ctype = str

    return ctype


def converttype():
    f = open(colfile,'r')
    fl = f.readlines()
    f.close()
    
    print "Converting",colfile
    cfile = colfile.replace(".txt","2.txt")

    fout = open(cfile,'w')
    for l in fl:
      cname = l.split()[0]
      ctype = str(" ".join(l.split()[1:]))
      nl = "%s %s" % (cname,gettype(ctype))
      print >> fout,nl
    fout.close()
    print "Wrote",cfile

#class PSC(IsDescription):
#   f = open(colfile,'r')
#   fl = f.readlines()
#   f.close()
#   for l in fl:
#     cname = l.split()[0]
#     ctype = str(" ".join(l.split()[1:]))
#     exec("%s = %s" % (cname,gettype(ctype)))

class PSC(IsDescription):
  ra = Float64Col()
  decl = Float64Col()
  err_maj = Float32Col()
  err_min = Float32Col()
  err_ang = Int16Col()
  designation = StringCol(17)
  j_m = Float32Col()
  j_cmsig = Float32Col()
  j_msigcom = Float32Col()
  j_snr = Float32Col()
  h_m = Float32Col()
  h_cmsig = Float32Col()
  h_msigcom = Float32Col()
  h_snr = Float32Col()
  k_m = Float32Col()
  k_cmsig = Float32Col()
  k_msigcom = Float32Col()
  k_snr = Float32Col()
  ph_qual = StringCol(3)
  rd_flg = StringCol(3)
  bl_flg = StringCol(3)
  cc_flg = StringCol(3)
  ndet = StringCol(6)
  prox = Float32Col()
  pxpa = Int16Col()
  pxcntr = Int64Col()
  gal_contam = Int16Col()
  mp_flg = Int16Col()
  pts_key = Int64Col()
  hemis = StringCol(1)
  date = StringCol(10)
  scan = Int16Col()
  glon = Float32Col()
  glat = Float32Col()
  x_scan = Float32Col()
  jdate = Float64Col()
  j_psfchi = Float32Col()
  h_psfchi = Float32Col()
  k_psfchi = Float32Col()
  j_m_stdap = Float32Col()
  j_msig_stdap = Float32Col()
  h_m_stdap = Float32Col()
  h_msig_stdap = Float32Col()
  k_m_stdap = Float32Col()
  k_msig_stdap = Float32Col()
  dist_edge_ns = Int64Col()
  dist_edge_ew = Int64Col()
  dist_edge_flg = StringCol(2)
  dup_src = Int16Col()
  use_src = Int16Col()
  a = StringCol(1)
  dist_opt = Float32Col()
  phi_opt = Int16Col()
  b_m_opt = Float32Col()
  vr_m_opt = Float32Col()
  nopt_mchs = Int16Col()
  ext_key = Int64Col()
  scan_key = Int64Col()
  coadd_key = Int64Col()
  coadd = Int16Col()
 
def listpars():
    cols = readcolnames()
    print "Coordinate parameters:"
    print "\t ra \t --> Right Ascension"
    print "\t decl \t --> Declination"
    print "\t glon \t --> Galactic Longitude"
    print "\t glat \t --> Galactic Latitude"
    print "\nFull list of parameters you can use for query:"
    print "\t%s" % (", ".join(cols))

def printdec(hname):
    extrema = lambda x: "[%f,%f]" % (float(min(x)),float(max(x)))

    bname = hname.split("/")[-1].replace(".h5","")
    group = bname[:5]
    tpth = "/%s/%s" % (group,bname)
    par = 'decl'

    h5file = openFile(hname, mode = "r")
    tbl = h5file.getNode(tpth)

    drange = "\"%s\" : %s\n" % (tbl._v_name,extrema(tbl.cols._f_col(par)[:]))

    sys.stdout.write(".");sys.stdout.flush()

    h5file.close()
    return drange

def printrange(par=None,tbl=None):
    median = lambda x: sorted(x)[int(len(x)/2)]
    extrema = lambda x: "[%f,%f,%f]" % (float(min(x)),float(max(x)),float(median(x)))
    hname = base+"/psc_all.h5" 
#    params = ['ra','decl','glon','glat']
    params = ['ra','decl']

    h5file = openFile(hname, mode = "r")
#    ptable = h5file.root.psc_a.psc_aaa
    pgroups = ["/psc_a","/psc_b"]
    ptables = lambda group: h5file.listNodes(group)
    fetchtbl = lambda t: h5file.getNode(t)
    tbl = ["/%s/%s" % (t[:5],t) for t in tbl]
    if tbl: 
       ptables = [fetchtbl(t) for t in tbl]
       [sys.stdout.write("%s\t\t%s %s\n" % (ptable._v_name,param,extrema(ptable.cols._f_col(param)[:]))) for ptable in ptables for param in params]
#    print h5file.__contains__(pgroup)
#    print h5file.root.psc_a._f_listNodes()
#    sys.exit()
    else:
      for pgroup in pgroups:
	for ptable in ptables:
#	    print "For table %s:" % (ptable._v_name)
#	    icols = readcolnames()
#	    if par and par in icols:
#	       print par,extrema(ptable.cols._f_col(par)[:])
#	    else:
	       [sys.stdout.write("%s\t\t%s %s\n" % (ptable._v_name,param,extrema(ptable.cols._f_col(param)[:]))) for param in params]
#    [sys.stdout.write("%s\t\t%s %s\n" % (ptable._v_name,param,extrema(ptable.cols._f_col(param)[:]))) for pgroup in pgroups for ptable in ptables(pgroup) for param in params]
    h5file.close()

def ddict():
    """dictionary of declination ranges contained
       in each psc file"""
    D = { "psc_aaa" : [-89.992844,-74.500000],
	  "psc_aab" : [-74.599998,-70.400002],
	  "psc_aac" : [-70.499992,-67.900002],
	  "psc_aad" : [-67.999992,-66.000000],
	  "psc_aae" : [-66.099991,-64.400002],
	  "psc_aaf" : [-64.499992,-63.100002],
	  "psc_aag" : [-63.199997,-62.000000],
	  "psc_aah" : [-62.099998,-61.000000],
	  "psc_aai" : [-61.099998,-59.900002],
	  "psc_aaj" : [-59.999996,-58.900002],
	  "psc_aak" : [-58.999996,-57.900002],
	  "psc_aal" : [-57.999996,-56.800018],
	  "psc_aam" : [-56.899998,-55.800003],
	  "psc_aan" : [-55.899998,-54.700001],
	  "psc_aao" : [-54.799999,-53.600002],
	  "psc_aap" : [-53.699997,-52.500000],
	  "psc_aaq" : [-52.599998,-51.400002],
	  "psc_aar" : [-51.499996,-50.200001],
	  "psc_aas" : [-50.299999,-49.000015],
	  "psc_aat" : [-49.099998,-47.900002],
	  "psc_aau" : [-47.999996,-46.700001],
	  "psc_aav" : [-46.799999,-45.500000],
	  "psc_aaw" : [-45.599998,-44.300003],
	  "psc_aax" : [-44.399998,-43.100002],
	  "psc_aay" : [-43.199997,-41.800007],
	  "psc_aaz" : [-41.899998,-40.600002],
	  "psc_aba" : [-40.699997,-39.300003],
	  "psc_abb" : [-39.399998,-38.100002],
	  "psc_abc" : [-38.199997,-36.900002],
	  "psc_abd" : [-36.999943,-35.600002],
	  "psc_abe" : [-35.699997,-34.400002],
	  "psc_abf" : [-34.499996,-33.300003],
	  "psc_abg" : [-33.399998,-32.100002],
	  "psc_abh" : [-32.199997,-31.000000],
	  "psc_abi" : [-31.099995,-29.800011],
	  "psc_abj" : [-29.899998,-28.800001],
	  "psc_abk" : [-28.899996,-27.600000],
	  "psc_abl" : [-27.699999,-26.500000],
	  "psc_abm" : [-26.599998,-25.300001],
	  "psc_abn" : [-25.399998,-24.100000],
	  "psc_abo" : [-24.199999,-22.900000],
	  "psc_abp" : [-22.999994,-21.600000],
	  "psc_abq" : [-21.699999,-20.400000],
	  "psc_abr" : [-20.499998,-19.100000],
	  "psc_abs" : [-19.199999,-17.800007],
	  "psc_abt" : [-17.899998,-16.500000],
	  "psc_abu" : [-16.599998,-15.100003],
	  "psc_abv" : [-15.199999,-13.700001],
	  "psc_abw" : [-13.799997,-12.300000],
	  "psc_abx" : [-12.399998,-10.800000],
	  "psc_aby" : [-10.899999,-9.300000],
	  "psc_abz" : [-9.399999,-7.800000],
	  "psc_aca" : [-7.899999,-6.300000],
	  "psc_acb" : [-6.399999,-4.800000],
	  "psc_acc" : [-4.899989,-3.200000],
	  "psc_acd" : [-3.299998,-1.500014],
	  "psc_ace" : [-1.599999,-0.000002],
	  "psc_baa" : [0.000000,1.699999],
	  "psc_bab" : [1.600001,3.399998],
	  "psc_bac" : [3.300000,5.099998],
	  "psc_bad" : [5.000000,6.799999],
	  "psc_bae" : [6.700000,8.499999],
	  "psc_baf" : [8.400000,10.199999],
	  "psc_bag" : [10.100000,11.899999],
	  "psc_bah" : [11.800001,13.699996],
	  "psc_bai" : [13.600002,15.399999],
	  "psc_baj" : [15.300000,17.099998],
	  "psc_bak" : [17.000000,18.899998],
	  "psc_bal" : [18.800001,20.699999],
	  "psc_bam" : [20.600000,22.499998],
	  "psc_ban" : [22.400000,24.299999],
	  "psc_bao" : [24.200003,26.199999],
	  "psc_bap" : [26.100000,28.099998],
	  "psc_baq" : [28.000004,29.999998],
	  "psc_bar" : [29.900002,31.999981],
	  "psc_bas" : [31.900000,33.899998],
	  "psc_bat" : [33.800003,35.899982],
	  "psc_bau" : [35.800003,37.899998],
	  "psc_bav" : [37.800003,39.999996],
	  "psc_baw" : [39.900002,42.099998],
	  "psc_bax" : [42.000000,44.299999],
	  "psc_bay" : [44.200005,46.499996],
	  "psc_baz" : [46.400002,48.599998],
	  "psc_bba" : [48.500000,50.799999],
	  "psc_bbb" : [50.700001,53.099998],
	  "psc_bbc" : [53.000000,55.299999],
	  "psc_bbd" : [55.200005,57.599998],
	  "psc_bbe" : [57.500000,60.099998],
	  "psc_bbf" : [60.000000,62.799999],
	  "psc_bbg" : [62.700001,66.199997],
	  "psc_bbh" : [66.100006,72.199974],
	  "psc_bbi" : [72.100006,89.990128],
	}
    return D

def makept(ifile,force=None):
    if not os.path.exists(ifile):
       print "%s does not exist!" % (ifile)
       print "Looking in %s..." % (ddir)
       ifile = "/".join([ddir,ifile])
       if not os.path.exists(ifile):
          print "%s does not exist!" % (ifile)
          sys.exit()
       else:
          print "Found %s" % (ifile)
    else:
       print "Found %s" % (ifile)

#    hname = base+"/psc_all.h5"
#    htitle = "All 2MASS PSC tables"
#    hmode = "a"
    hname = "/".join([base,ifile.replace(".gz",".h5").split("/")[-1]])
    bname = ifile.replace(".gz","")
    if "/" in bname:
       bname = bname.split("/")[-1]
    tdesc = "table for %s" % (bname)
    htitle = "HDF5 file for %s" % (bname)
    hmode = "w"
    if ifile.split("_")[1][0] == "a":
       gname = "psc_a"
       gdesc = "PSC group A"
    elif ifile.split("_")[1][0] == "b":
       gname = "psc_b"
       gdesc = "PSC group B"

    h5file = openFile(hname, mode = hmode, title = htitle)
    print
#    print h5file
    if not h5file.__contains__("/"+gname):
       group = h5file.createGroup("/", gname, gdesc)
    else:
       group = "/"+gname
#    print group.__contains__("/"+bname)
#    print h5file.listNodes(group)
    tpath = "%s/%s" % (group,bname)
    if not h5file.__contains__(tpath):
      table = h5file.createTable(group, bname, PSC, tdesc)
    elif force:
      h5file.removeNode(tpath)
      print "Removed %s" % (tpath)
      table = h5file.createTable(group, bname, PSC, tdesc)
    else:
      print "A PyTable for %s already exists!" % (bname)
      sys.exit()

    pfile = readpsc(ifile)

    print "%s has %d objects." % (ifile,len(pfile))
    #print "pfile is of type: ",type(pfile)
    print "%s has %d columns" % (ifile,len(pfile[0].split("|")))
    #print "the first 10 entries of the first row are: ",pfile[0].split("|")[:10]
    #print "\npfile[0] = ",pfile[0]

    cols = readcolnames()
#    print "There are %d column names" % (len(cols))
#    print cols


    print "Creating PyTable for %s..." % (bname)
    prow = table.row
    for l in pfile:
      fields = l.split("|")
      for col,field in zip(cols,fields):
        if "N" in field:
          field = "-99"
          #print col,field
        prow[col] = field
      prow.append()
    print "Wrote %s" % (hname)
    print
    print h5file

    table.flush() #flush table's I/O buffer
    h5file.close()

def querypt(par1,prange1,par2=None,prange2=None,hname=None,ofile=None,mkdict=0):
    if mkdict: verb = 0
    else: verb = 1
    if hname:
       if not os.path.exists(hname):
          print "%s does not exist!" % (hname)
          print "Looking elsewhere..."
          if os.path.exists(base+"/"+hname):
             hname = base+"/"+hname
             print "Found %s" % (hname)
          elif os.path.exists(ddir+"/"+hname):
             hname = ddir+"/"+hname
             print "Found %s" % (hname)
          else:
             print "%s does not exist!" % (hname)
             sys.exit()
       elif ".h5" not in hname:
          print "%s is not an .h5 file! Please specify a file in HDF format."
          sys.exit()
    if not par1 or not prange1:
       print "You must specify a parameter and a range to use in query!"
       help()

    if ofile: fname = ofile

    spec = 0
    if hname: spec = 1

    range1 = [float(v) for v in prange1.split(",")]
    if prange2: range2 = [float(v) for v in prange2.split(",")]

    if spec:
       tnames = ['%s' % (hname.replace(".h5",""))]
    else:
       D = ddict()
       dk = sorted(D.keys())
       dind = lambda x: dk.index(x)
       if par1 == 'decl': decrange = range1
       elif par2 == 'decl': decrange = range2
#       deca,decb = decrange
#       print deca,decb
#       if abs(deca-decb) > 1.0:
#          dsteps = range(int(deca),int(decb)+1,1)
#          tname = [key for num in dsteps for key in D.keys() if D[key][0] <= num <= D[key][1]]
#       else:
       tname = [key for num in decrange for key in D.keys() if D[key][0] <= num <= D[key][1]]
       if abs(dind(tname[0]) - dind(tname[-1])) > 1:
          dsteps = range(dind(tname[0]),dind(tname[-1])+1)
	  tname = [dk[i] for i in dsteps]
       tnames = []
       [tnames.append(t) for t in tname if t not in tnames]

    icols = readcolnames()
    if ofile:
#       fname = base+"/"+fname
       fout = open(fname,"w")
       head = "# %s" % ("|".join(icols))
       print >> fout,head

    if mkdict: 
       data = {}
       for icol in icols:
         data[icol] = []

    if verb:
       print "Using the parameter '%s' with a range of %f to %f" % (par1,range1[0],range1[1])
       if prange2: 
          print "Using the parameter '%s' with a range of %f to %f" % (par2,range2[0],range2[1])

    for tname in tnames:
        group = tname[:5]
        tpth = "/%s/%s" % (group,tname)
#        print tpth
        if not spec: hname = "%s/%s.h5" % (base,tname)
        if not os.path.exists(hname):
           print "%s does not exist!" % (hname)
           sys.exit()
        h5file = openFile(hname, mode = "r")
#        ptable = h5file.root.psc_a.psc_aaa
        ptable = h5file.getNode(tpth)

#        head = "# %s" % ("\t".join(icols))
        n = 0
        print "Sifting through %s table...\n" % (tname),;sys.stdout.flush()

	icols = ["%s" % (str(icol)) for icol in icols]
        if not prange2:
           for x in ptable.where("""(%f < %s) & (%s < %f)""" \
           % (range1[0],par1,par1,range1[1])):
               n += 1
               irows = []
	       if mkdict: 
	          [data[icol].append(gettype2(icol)(x[icol])) for icol in icols]
	       if ofile:
                  [irows.append(str(x[icol])) for icol in icols]
                  iline = "|".join(irows)
                  print >> fout,iline
#               if n and not (n % 1000): sys.stdout.write(".");sys.stdout.flush()
#               if n and not (n % 10): 
	       sys.stdout.write("%d objects found\r" % (n));sys.stdout.flush()

        else:
           for x in ptable.where("""(%f < %s) & (%s < %f) & \
           (%f < %s) & (%s < %f)""" % (range1[0],par1,par1,range1[1], \
           range2[0],par2,par2,range2[1])):
#           for x in ptable.iterrows():
#             if range1[0] < x[par1] < range1[1]:
#             print "good",;sys.stdout.flush()
#             print x['ra'],x['decl']
#             sys.exit()
#             if range2[0] < x[par2] < range2[1]:
               n += 1
               irows = []
	       if mkdict: 
	          [data[icol].append(gettype2(icol)(x[icol])) for icol in icols]
	       if ofile:
                  [irows.append(str(x[icol])) for icol in icols]
                  iline = "|".join(irows)
                  print >> fout,iline
#               if n and not (n % 1000): sys.stdout.write(".");sys.stdout.flush()
#               if n and not (n % 10): 
	       sys.stdout.write("%d objects found\r" % (n));sys.stdout.flush()

        print "%d objects found" % (n)
        h5file.close()

    if mkdict: 
       for icol in icols:
         if gettype2(icol) != "str":
            data[icol] = asarray(data[icol])
       return data

    if ofile: 
       fout.close()
       print "Wrote",fname

def help():
    print "--------------------------------------------------"
    print "Current possible options:"
    print "  -c \t\t\t--> convert PSC column file"
    print "  -f <file> \t\t--> file to generate PyTable from"
    print "  -q [<file>] \t\t--> query all PSC .h5 files; if <file> is specified,"
    print "   \t\t\t    search will only be performed in that file"
    print "  -o <file> \t\t--> name of file to output query"
    print "  -p1 <parameter> \t--> name of (first) parameter to use in query"
    print "  -r1 <lower>,<upper> \t--> lower,upper values of (first) <parameter>" 
    print "  -p2 <parameter> \t--> (optional) name of second parameter to use in query"
    print "  -r2 <lower>,<upper> \t--> (optional) lower,upper values of second <parameter>" 
    print "   \t\t\t    to use in query"
    print "  -h \t\t\t--> help; list options, usage"
    print "  -lp \t\t\t--> list of parameters usable for query"
    print "  -pr \t\t\t--> print coordinate range of a particular table"
    print
    print "Examples:"
    print "  %s -f psc_aaa.gz" % (sys.argv[0])
    print "  %s -q psc_aaa.h5 -o twomass_cut.txt -p1 ra -r1 273.0,274.0 -p2 decl -90.0,-89.0" % (sys.argv[0])
    print "--------------------------------------------------"
    sys.exit()

if __name__ == "__main__":
   for n,arg in enumerate(sys.argv):
      if arg == "-f": 
         ifile = sys.argv[n+1]
	 if "," in ifile:
	    ifile = ifile.split(",")
      elif "-c" in arg: convert = 1
      elif "-all" in arg: all = 1
      elif "-force" in arg: force = 1
      elif "-q" in arg: 
	if ".h5" in sys.argv[n+1]:
           hname = sys.argv[n+1]
	query = 1
      elif "-o" in arg: ofile = sys.argv[n+1]
      elif "-p1" in arg: par1 = sys.argv[n+1]
      elif "-p2" in arg: par2 = sys.argv[n+1]
      elif "-r1" in arg: prange1 = sys.argv[n+1]
      elif "-r2" in arg: prange2 = sys.argv[n+1]
      elif "-lp" in arg: lp = 1
      elif "-pr" in arg: pr = 1
      elif "-tbl" in arg: 
        tbl = sys.argv[n+1]
	if "," in tbl: tbl = tbl.split(",")
	else: tbl = [tbl]
      elif "-h" in arg: 
        help()

   if len(sys.argv) < 2:
      print "Please specify an option!\n"
      help()

   if query:
      if not ofile:
         print "You didn't specify a file to output query to!"
         ofile = "table_query.txt"
         q1 = "Will write query to %s. Is this okay? (y/n): " % (ofile)
         a1 = raw_input(q1)
         if a1 == 'n':
            q2 = "Please type name of output file: "
            ofile = str(raw_input(q2))
      querypt(par1,prange1,par2=par2,prange2=prange2,hname=hname,ofile=ofile,mkdict=mkdict)

   if convert:
      converttype() 

   if ifile:
      if type(ifile).__name__ == 'str':
         makept(ifile,force=force)
      elif type(ifile).__name__ == 'list':
         [makept(f,force=force) for f in ifile]

   if all:
      plist = glob(ddir+"/psc_???.gz")
      plist = [f for f in plist if not os.path.exists("/".join([base,f.split("/")[-1].replace(".gz",".h5")]))]
      [makept(f) for f in plist]

   if lp:
      listpars()

   if pr:
#      printrange(tbl=tbl)
      hlist = glob(base+"/psc_???.h5")
      dlname = "dec_list.txt"
      print "Writing",dlname
      f = open(dlname,"w")
      [f.write(printdec(hf)) for hf in hlist]
      f.close()
      print "\nWrote",dlname

   printtime()


